"""
This sub package includes multiple modules of test examples to check if the
language server can deal with a complex MRO list where all the relevant class
declarations are distributed in different files.

The example is taken from the [Wikipedia C3 linearization page](
https://en.wikipedia.org/wiki/C3_linearization). The original author keeps the
copyright and please refer to the wiki page for more details.
"""